@@ -20,8 +20,9 @@ from datetime import datetime
 from tensorboard.backend.event_processing.event_accumulator import EventAccumulator
 import warnings
 warnings.filterwarnings(action='ignore')
-from pytorch_lightning.utilities.seed import seed_everything as pl_set_seed
-
+import shutil
+from pytorch_lightning.callbacks import RichProgressBar, TQDMProgressBar
+import random
 
 class BaseExperiment:
     """
@@ -123,7 +124,7 @@ class BaseExperiment:
         y_test : Optional[np.ndarray] = None,
         test_size : float = 0.01,
         stack_models : bool = True,
-        random_state : int = 0,
+        random_number : int = 0,
         path : Optional[str] = None
     ) -> None:
         ##########################################
@@ -139,7 +140,7 @@ class BaseExperiment:
         self.y_test = y_test
         self.test_size = test_size
         self.stack_models = stack_models
-        self.random_state = random_state
+        self.random_number = random_number
         self.path = "." if path is None else path
         if self.path[-1] == "/":
             self.path = self.path[:-1]
@@ -171,19 +172,20 @@ class BaseExperiment:
         self.project_path = self.path + "/" + self.project_name 
         self._create_folder(self.project_path)
         
-
+        self.saved_model_types = {}
         self.exp_path = None
         self.log_path = None
 
         self.n_ckpts = 0
-        self.exp_meta = {}
+        self.exp_metadata = {}
         now = datetime.now()
         self.birthday = now.strftime("%b-%d-%Y-%H-%M-%S")
 
         
         self.exp_metadata = {'date of birth of this experiemt' : self.birthday}
-        pl_set_seed(self.random_state)
-        np.random.seed(self.random_state)
+
+        self.rng = np.random.default_rng(random_number)
+        self.random_state = int(10**6 * self.rng.random(1))
 
 
     def update_log_path(self):
@@ -191,16 +193,16 @@ class BaseExperiment:
         self.log_path = self.exp_path + "/logs"
 
     def update_exp_metadata(self):
-        self.exp_meta.update({'date of birth of this experiemt' : str(self.birthday)})
-        self.exp_meta.update({'# of checkpoints saved' : int(self.n_ckpts)})
-        self.exp_meta.update({'size of the experiment in MB' : float(os.path.getsize(self.exp_path)/10**6)})
-        self.exp_meta.update({'size of stack in MB': float(sys.getsizeof(self.stack)/10**6)})
-        self.exp_meta.update({'random state' : int(self.random_state)})
+        
+        self.exp_metadata.update({'# of checkpoints saved' : int(self.n_ckpts)})
+        self.exp_metadata.update({'size of the experiment in MB' : float(os.path.getsize(self.exp_path)/10**6)})
+        self.exp_metadata.update({'size of stack in MB': float(sys.getsizeof(self.stack)/10**6)})
+        self.exp_metadata.update({'random state' : int(self.random_state)})
         if self.data_module:
-            self.exp_meta.update({'size of data in MB' : float(sys.getsizeof(self.X)/10**6)})
+            self.exp_metadata.update({'size of data in MB' : float(sys.getsizeof(self.X)/10**6)})
         json_name = self.exp_path +"/exp_metadata.json" 
         with open(json_name, 'w') as fout:
-            json.dump(self.exp_meta, fout)
+            json.dump(self.exp_metadata, fout)
     # *********************************************************************        Class Methods        ********************************************************************* 
     #######################################################################  set_model (Begins) #######################################################################
    
@@ -213,8 +215,6 @@ class BaseExperiment:
         model_type,
         **hparams : Any
     ):
-        
-        np.random.seed(self.random_state)
         if BaseEstimator in inspect.getmro(model_type):
             self.model = model_type(**hparams)
             self.model_metadata = self.set_model_metadata(self.model, hparams)
@@ -237,7 +237,6 @@ class BaseExperiment:
         if self.model is None:
             raise AssertionError("Model must be defined first. Use .model(model_type, **hparams) to declare a model.")
         
-        np.random.seed(self.random_state)
         del self.data_module
         self.data_module = None
         if isinstance(self.model, ConvNet):
@@ -341,22 +340,28 @@ class BaseExperiment:
         self.update_log_path()
         self._create_folder(self.log_path)
         trainer_setting = trainer_config
-        
+        if self.logger:
+            del self.logger
+        self.logger = pl_loggers.TensorBoardLogger(save_dir=self.log_path+"/tensorboardlogs/" + self.model.__class__.__name__, name=self.model_metadata)
+
         if torch.cuda.is_available():
             trainer_setting.update({'gpus' : -1})
             trainer_setting.update({'accelerator' : 'gpu'})
         if 'min_epochs' not in trainer_config.keys():
-            trainer_setting.update({'min_epochs' : 5})
+            trainer_setting.update({'min_epochs' : 10})
         if 'max_epochs' not in trainer_config.keys():
-            trainer_setting.update({'max_epochs' : 30})
+            trainer_setting.update({'max_epochs' : 150})
         if 'sample_weight' in trainer_setting.keys():
             trainer_setting.pop('sample_weight')
-        if self.logger:
-            del self.logger
-        self._create_folder(self.log_path+"/tensorboardlogs/"+self.model_metadata)
-        self.logger = pl_loggers.TensorBoardLogger(save_dir=self.log_path+"/tensorboardlogs", name=self.model_metadata)
-
-        return pl.Trainer(logger=self.logger, **trainer_config)
+        if 'callbacks' not in trainer_setting.keys():
+            trainer_setting.update({'callbacks' : []})
+        trainer = pl.Trainer(logger=self.logger, **trainer_config)
+        for c in trainer.callbacks:
+            if isinstance(c, TQDMProgressBar):
+                trainer.callbacks.remove(c)
+                trainer.callbacks.append(RichProgressBar())
+        self._create_folder(self.log_path+"/tensorboardlogs/"+ self.model.__class__.__name__ + "/" + self.model_metadata)
+        return trainer
 
     #######################################################################  fit (Begins) #######################################################################
     """
@@ -393,7 +398,7 @@ class BaseExperiment:
     ):
         if self.model is None:
             raise AssertionError("Model must be defined first. Use .model(model_type, **hparams) to declare a model.")
-        np.random.seed(self.random_state)
+        
         self.fit_or_cv = fit_or_cv
         self.update_log_path()
         self.data_module = self.get_data_module()
@@ -401,7 +406,6 @@ class BaseExperiment:
             self.logger.close()
             del self.logger
             self.logger = None
-
         if BaseEstimator in inspect.getmro(self.model.__class__):
             return self.fit_sklearn(sample_weight=sample_weight, **config)
         elif pl.LightningModule in inspect.getmro(self.model.__class__):
@@ -414,12 +418,9 @@ class BaseExperiment:
         sample_weight : Optional["np.ndarray"] = None
     ):
         # for consistency with the logging style of pytorch_lightning.
-        folder_dir = self.log_path + "/tensorboardlogs" 
-        self._create_folder(folder_dir)
+        folder_dir = self.log_path + "/tensorboardlogs/" + self.model.__class__.__name__ 
         version = len(os.listdir(folder_dir))
-        self._create_folder(folder_dir + "/" + self.model_metadata + "/" + "version_" + str(version))
-        log_dir = folder_dir + "/" + self.model_metadata + "/" + "version_" + str(version)
-
+        log_dir = folder_dir + "/" + self.model_metadata + "/" + "version_{}".format(version)
         self.logger = SummaryWriter(log_dir=log_dir)
         self.model.fit(self.X_train, self.y_train, sample_weight)
         y_train_preds = self.model.predict(self.X_train)
@@ -513,8 +514,7 @@ class BaseExperiment:
         if trainer is None and isinstance(model, pl.LightningModule):
             trainer = self.trainer  
         self.update_log_path()
-        checkpoint_path = self.log_path + "/checkpoints"
-        self._create_folder(checkpoint_path)
+        
         if isinstance(model, BaseEstimator):
             model_metadata = self.set_model_metadata(model, model.__dict__)
             ckpt_name = model_metadata + ".sav"
@@ -523,18 +523,27 @@ class BaseExperiment:
             ckpt_name = model_metadata + ".ckpt"
         else:
             raise NotImplementedError("Astrape v 0.0.0 only supports variants of sklearn and pl.LightningModule.")
+        checkpoint_path = self.log_path + "/checkpoints/" + model.__class__.__name__ + "/" + model_metadata
+        self._create_folder(checkpoint_path)
         ckpt_path = checkpoint_path +"/" + ckpt_name 
         
         if isinstance(model, BaseEstimator):
             pickle.dump(model, open(ckpt_path, 'wb'))
         elif isinstance(model, pl.LightningModule):
             trainer.save_checkpoint(ckpt_path)
-        json_dir = self.log_path + "/jsonlogs"
-        self._create_folder(json_dir)
-        tensorboard_logdir = self.logger.experiment.get_logdir()
 
+        metadata_dir = self.log_path + "/jsonlogs/" + model.__class__.__name__ + "/" + model_metadata
+        self._create_folder(metadata_dir)
+        version = len(os.listdir(metadata_dir))
+        self._create_folder(metadata_dir+"/version_{}".format(version))
+        tensorboard_logdir = self.logger.experiment.get_logdir()
+        if model.__class__ in self.saved_model_types:
+            self.saved_model_types.update({model.__class__: self.saved_model_types[model.__class__] + 1})
+        else:
+            self.saved_model_types.update({model.__class__: 1})
+        
         self.n_ckpts += 1 
-        self.log_metrics_to_json(model, tensorboard_logdir, val_metrics)
+        self.log_metrics_to_json(model, tensorboard_logdir, version, val_metrics)
         self.update_exp_metadata()
     
     #######################################################################  save_ckpt (Ends)  ####################################################################### 
@@ -608,8 +617,37 @@ class BaseExperiment:
             self.save_ckpt(**subject)
         del self.stack 
         self.stack = []
-    #######################################################################  save_stack (Ends)  ####################################################################### 
 
+    def delete_saved_model(
+        self,
+        model_type : "pl.LightningModule",
+        **hparams
+    ):
+        pseudomodel = model_type(**hparams)
+        model_metadata = self.set_model_metadata(pseudomodel, hparams)
+        ckpt_dir = self.log_path + '/checkpoints/' + model_type.__name__ + "/" + model_metadata
+        jsonlog_dir = self.log_path + '/jsonlogs/' + model_type.__name__ + "/" + model_metadata
+        tensorboard_dir = self.log_path + '/tensorboardlogs/' + model_type.__name__ + "/" + model_metadata
+        if not os.path.exists(ckpt_dir) or not os.path.exists(jsonlog_dir) or not os.path.exists(tensorboard_dir):
+            return None
+        
+
+        success = {ckpt_dir:0, jsonlog_dir:0, tensorboard_dir:0}
+        if os.path.exists(ckpt_dir):
+            shutil.rmtree(ckpt_dir)
+            success.update({ckpt_dir:1})
+        if os.path.exists(jsonlog_dir):
+            shutil.rmtree(jsonlog_dir)
+            success.update({jsonlog_dir:1})
+        if os.path.exists(tensorboard_dir):
+            shutil.rmtree(tensorboard_dir)
+            success.update({tensorboard_dir:1})
+
+        for key, value in success.items():
+            if value == 0:
+                print("There is no directory {}".format(key))
+            
+        
     #######################################################################  best_ckpt_thus_far (Begins)  ####################################################################### 
     """
     Method for returning the best saved checkpoint(or the path name of the best saved checkpoint) from the directory. The criterion is the type of the validation metric.
@@ -624,7 +662,7 @@ class BaseExperiment:
     """
     def best_ckpt_thus_far(self, val_metric : str = "val/acc", fetch_only_dir : bool = False):
         try:
-            jsonlogdir = self.log_path + "/jsonlogs/"
+            jsonlogdir = self.log_path + "/jsonlogs"
         except:
             print("Nothing has been saved yet.")
             return None
@@ -632,26 +670,37 @@ class BaseExperiment:
         best_val = None
         ckpt_extension_type = None
         model_type = None
-        for file in os.listdir(jsonlogdir):
-            with open(jsonlogdir + "/" + file, "r") as json_file:
-                logs = json.load(json_file)
-                val_score = logs[val_metric] if isinstance(logs[val_metric], float) else logs[val_metric][-1]
-                if best_val is None:
-                    best_val = val_score
-                    best_path = file
-                    model_type = logs['model_type']
-                elif mode == "max" and val_score > best_val:
-                    best_val = val_score
-                    best_path = file
-                    model_type = logs['model_type']
-                elif mode == "min" and val_score < best_val:
-                    best_val = val_score
-                    best_path = file
-                    model_type = logs['model_type']
-                info = {val_metric : best_val}
+        best_model_metadata = None
+        best_version = None
+        for model_type_dir in os.listdir(jsonlogdir):
+            for metadata in os.listdir(model_type_dir):
+                for version in os.listdir(metadata):
+                    for file in os.listdir(version):
+                        with open(model_type_dir + "/" + version + "/" + file, "r") as json_file:
+                            logs = json.load(json_file)
+                            val_score = logs[val_metric] if isinstance(logs[val_metric], float) else logs[val_metric][-1]
+                            if best_val is None:
+                                best_val = val_score
+                                best_path = file
+                                model_type = logs['model_type']
+                                best_model_metadata = metadata
+                                best_version = version
+                            elif mode == "max" and val_score > best_val:
+                                best_val = val_score
+                                best_path = file
+                                model_type = logs['model_type']
+                                best_model_metadata = metadata
+                                best_version = version
+                            elif mode == "min" and val_score < best_val:
+                                best_val = val_score
+                                best_path = file
+                                model_type = logs['model_type']
+                                best_model_metadata = metadata
+                                best_version = version
+                            info = {val_metric : best_val}
         if not fetch_only_dir:
-            ckpt_path = self.log_path + '/checkpoints/'
-            ckpt_name = ckpt_path + str(best_path).replace(".json", "")
+            ckpt_path = self.log_path + '/checkpoints/' + model_type + "/" + best_model_metadata + "/" + best_version
+            ckpt_name = ckpt_path + "/" + str(best_path).replace(".json", "")
             extension_candidates = [".ckpt", ".sav"]
             for extension_type in extension_candidates:
                 ckpt = ckpt_name + extension_type
@@ -667,8 +716,8 @@ class BaseExperiment:
             
             return best_ckpt, info
         else:
-            ckpt_path = self.log_path + '/checkpoints/'
-            ckpt_name = ckpt_path + str(best_path).replace(".json", "")
+            ckpt_path = self.log_path + '/checkpoints/' + model_type + "/" + best_model_metadata + "/" + best_version
+            ckpt_name = ckpt_path + "/" +str(best_path).replace(".json", "")
             extension_candidates = [".ckpt", ".sav"]
             for extension_type in extension_candidates:
                 ckpt = ckpt_name + extension_type
@@ -785,14 +834,16 @@ class BaseExperiment:
         self, 
         model,
         tensorboard_logdir: str,
+        version : int,
         metrics_dict : Optional[dict] = None
     ) -> None:
         log = {'model_type' : model.__class__.__name__}
         error_count = 0
         self.update_log_path()
-        json_log_dir = self.log_path + "/jsonlogs"
         model_metadata = self.set_model_metadata(model, self.get_hparams(model))
-        json_logdir = json_log_dir + "/" + model_metadata + ".json"
+        metadata_dir_ = self.log_path+'/jsonlogs/'+ model.__class__.__name__ + "/" + model_metadata
+        jsonlog_dir = metadata_dir_ + "/" + 'version_{}'.format(version)
+        json_logdir = jsonlog_dir + "/validation_metrics.json"
 
         if not metrics_dict:
             event = EventAccumulator(tensorboard_logdir)
